type Fn<T extends HTMLElement> = (el: Component<T>) => void | Promise<void>

/**
 * Client-side Component management class
 *
 * just add a `data-component="name"` to your component to make it work flowlessly
 */
export default class Component<T extends HTMLElement> {
	private static components: Record<string, (el: HTMLElement) => void | Promise<void>> = {}
	private constructor(
		public element: T
	) {}

	public handled(value: boolean): this
	public handled(): boolean
	public handled(value?: boolean): this | boolean {
		if (typeof value === 'undefined') {
			return typeof this.element.dataset.handled === 'string'
		}
		this.element.dataset.handled = ''
		return this
	}

	public init(fn: (el: Component<T>) => void | Promise<void>) {
		if (this.handled()) {
			return
		}
		fn(this)
		this.handled(true)
	}

	public child<El extends HTMLElement>(query: string, fn: Fn<El>) {
		this.element.querySelectorAll<El>(query).forEach((it) => {
			const cp = new Component(it)
			cp.init(fn)
		})
	}

	/**
	 * start handling an element
	 * @param query the query to get the element
	 * @param fn the function that is run ONCE per elements
	 */
	public static handle<T extends HTMLElement>(query: string, fn: (el: T) => void | Promise<void>) {
		document.querySelectorAll<T>(query).forEach((it) => {
			const cp = new Component(it)
			cp.init((it) => fn(it.element))
		})
		document.addEventListener('astro:page-load', () => {
			document.querySelectorAll<T>(query).forEach((it) => {
				const cp = new Component(it)
				cp.init((it) => fn(it.element))
			})
		})
	}

	public static addComponent<T extends HTMLElement>(name: string, fn: (el: T) => void | Promise<void>) {
		this.components[name] = fn as any
		this.load()
	}

	public static load(base?: HTMLElement) {
		if (base) {
			this.loadElement(base)
		}
		(base ?? document.body).querySelectorAll<HTMLElement>('[data-component]').forEach((it) => {
			this.loadElement(it)
		})
	}

	private static loadElement(it: HTMLElement) {
		const component = it.dataset.component
		if (!component) {
			return
		}
		const fn = this.components[component]
		if (!fn) {
			return
		}
		fn(it)
		it.removeAttribute('data-component')
	}
}
