import type { ImportMetaEnv } from 'env'

// eslint-disable-next-line @typescript-eslint/no-unused-vars
const defaults: Array<keyof ImportMetaEnv> = [
	'POSTGRES_USERNAME',
	'POSTGRES_PASSWORD',
	'POSTGRES_DATABASE',
	'POSTGRES_HOST',
	'POSTGRES_PORT'
]

/**
 * Get the environment variable
 *
 * @param key the env variable key
 * @param defaultValue a default value if applicable
 * @returns the environment value or undefined if not found
 */
export function getEnv(key: keyof ImportMetaEnv, defaultValue: string): string
export function getEnv(key: keyof ImportMetaEnv, defaultValue?: string | undefined): string | undefined
export function getEnv(key: keyof ImportMetaEnv, defaultValue?: string | undefined): string | undefined {
	// get the env variable through Astro > NodeJS > input
	const res = import.meta.env[key] ?? process.env[key] ?? defaultValue

	return res ?? undefined
}

/**
 * get the environment variable and throws if not found
 *
 * @throws {Error} if the env variable is not found
 * @param key the env variable key
 * @param defaultValue a default value if applicable
 * @returns the environment value
 */
export function requireEnv(key: keyof ImportMetaEnv, defaultValue?: string): string {
	// get the env variable through Astro > NodeJS > input
	const res = getEnv(key, defaultValue)

	// throw if env variable is not set
	if (!res) {
		throw new Error(`MissingEnvError: the env ${key} is not set!`)
	}
	return res
}

export function envExists(key: keyof ImportMetaEnv): boolean {
	return !!(import.meta.env[key] ?? process.env[key])
}
