import type Schema from 'libs/Schema'
import type { SchemaInfer } from 'libs/Schema'
import type { Query } from '../Query'

export interface DBPull<T extends Schema> {
	/**
	 * total number of rows that are valid with the specified query
	 */
	rows: number
	/**
	 * total number of rows in the table
	 */
	rowsTotal: number

	/**
	 * current page number
	 */
	page: number

	/**
	 * total amount of pages
	 */
	pageTotal: number

	/**
	 * the data fetched
	 */
	data: Array<SchemaInfer<T>>
}

/**
 * the Dao is the object that connect the Database or source to the application layer
 *
 * you MUST call it through the `DaoFactory` file
 */
export default interface DaoAdapter<T extends Schema> {
	/**
	 * create a new object in the remote source
	 *
	 * @param obj the object to create
	 */
	create?(obj: Partial<SchemaInfer<T>>): Promise<SchemaInfer<T> | null>

	/**
	 * read from the remote source
	 *
	 * @param query the query to filter/sort results
	 */
	read?(query?: Query<SchemaInfer<T>>): Promise<DBPull<T>>

	/**
	 * update an object to the remote source
	 *
	 * @param obj the object to update
	 */
	update?(obj: Partial<SchemaInfer<T>>): Promise<SchemaInfer<T> | null>

	/**
	 * depending if the object already exists or not
	 * it will update an exisintg object or create a new one
	 *
	 * @param obj the object to insert/update
	 */
	upsert?(obj: Partial<SchemaInfer<T>>): Promise<SchemaInfer<T> | null>

	/**
	 * update an object to the remote source
	 *
	 * @param id (DEPRECATED) the ID of the object
	 * @param obj the object to patch (MUST include ids, and changes)
	 */
	patch?(id: string, obj: Partial<SchemaInfer<T>>): Promise<SchemaInfer<T> | null>

	/**
	 * delete an object from the source
	 * @param obj the object ot delete (it must at least include the id(s))
	 */
	delete?(obj: Partial<SchemaInfer<T>>): Promise<boolean>
}
