import type Schema from 'libs/Schema'
import type { SchemaInfer } from 'libs/Schema'
import type DaoAdapter from './Adapters/DaoAdapter'
import type { DBPull } from './Adapters/DaoAdapter'
import type { Query } from './Query'

/**
 * the Dao is the object that connect the Database or source to the application layer
 *
 * you MUST call it through the `DaoFactory` file
 */
export default class Dao<S extends Schema = Schema> {

	public constructor(
		public readonly schema: S,
		public readonly adapter: DaoAdapter<S>
	) { }

	/**
	 * insert a new object into the source
	 *
	 * @param obj the object to create
	 * @returns the object with it's id filled if create or null otherwise
	 */
	public async create(obj: Partial<SchemaInfer<S>>): Promise<SchemaInfer<S> | null> {
		if (!this.adapter.create) {
			throw new Error('the Adapter does not allow you to create elements')
		}
		return this.adapter.create(obj)
	}

	/**
	 * insert a new object into the source
	 *
	 * @param obj the object to create
	 * @returns the object with it's id filled if create or null otherwise
	 */
	public async insert(obj: Partial<SchemaInfer<S>>): Promise<SchemaInfer<S> | null> {
		return this.create(obj)
	}

	/**
	 * find the list of objects having elements from the query
	 *
	 * @param query a partial object which filter depending on the elements, if not set it will fetch everything
	 * @returns an array containing the list of elements that match with the query
	 */

	public async findAll(query?: Query<SchemaInfer<S>>): Promise<DBPull<S>> {
		if (!this.adapter.read) {
			throw new Error('the Adapter does not allow you to read from the remote source')
		}
		return this.adapter.read(query)
	}

	/**
	 * find the list of objects having elements from the query
	 *
	 * @param query a partial object which filter depending on the elements, if not set it will fetch everything
	 * @returns an array containing the list of elements that match with the query
	 */
	public async find(query: Parameters<this['findAll']>[0]) {
		return this.findAll(query)
	}

	/**
	 * find an object by it's id
	 *
	 * (shortcut to findOne({id: id}))
	 *
	 * @param id the id of the object
	 * @returns
	 */
	public findById(id: any): Promise<SchemaInfer<S> | null> {
		return this.findOne({ id: id })
	}

	/**
	 * find an object by it's id
	 *
	 * (shortcut to findOne({id: id}))
	 *
	 * @param id the id of the object
	 * @returns
	 */
	public get(id: any) {
		return this.findById(id)
	}

	/**
	 * find the first element that match `query`
	 *
	 * @param query a partial object which filter depending on the elements, if not set it will fetch everything
	 * @returns the first element matching with the query or null otherwise
	 */
	public async findOne(query?: Parameters<this['findAll']>[0]): Promise<SchemaInfer<S> | null> {
		return (await this.findAll(query)).data[0] ?? null
	}

	/**
	 * update the remote reference of the object
	 *
	 * note: it will not try to insert an item (use `upsert` to handle this)
	 *
	 * @param obj the object to update
	 * @returns an object if it was able to update or null otherwise
	 */
	public async update(obj: Partial<SchemaInfer<S>>): Promise<SchemaInfer<S> | null> {
		if (!this.adapter.update) {
			throw new Error('the Adapter does not allow you to update to the remote source')
		}
		return this.adapter.update(obj)
	}

	/**
	 * change some elements from the object and return the object updated
	 * @param id the id of the object
	 * @param changes the change to make
	 */
	public async patch(id: string, changes: Partial<SchemaInfer<S>>): Promise<SchemaInfer<S> | null> {
		if (!this.adapter.patch) {
			const query = await this.findById(id)
			if (!query) {
				return null
			}
			return await this.update({ ...query, ...changes })
		}
		return await this.adapter.patch(id, changes)
	}

	/**
	 * update the remote reference of the object or create it if not found
	 * @param obj the object to update/insert
	 * @returns the object is updated/inserted or null otherwise
	 */
	public async upsert(object: Partial<SchemaInfer<S>>): Promise<Partial<SchemaInfer<S>> | null> {
		if (!this.adapter.upsert) {
			throw new Error('the Adapter does not allow you to upsert to the remote source')
		}
		return this.adapter.upsert(object)
	}

	/**
	 * Delete the object
	 * @param obj the object or ID to delete
	 *
	 * @returns if the object was deleted or not (if object is not in db it will return true)
	 */
	public async delete(obj: Partial<SchemaInfer<S>>): Promise<boolean> {
		if (!this.adapter.delete) {
			throw new Error('the Adapter does not allow you to delete on the remote source')
		}
		return this.adapter.delete(obj)
	}
}
